/*
	Ceci est une modification de ttstart pour dcompresser un programme ppg oncalc
	Ce programme est bas sur la version 1.17 de ttstart (incluse dans tt v1.31)
	Modifications du source de la TIGCC Tools Suite :
	- Retrait des fonctions d'excution
	- Modification de l'allocation de la sortie (fichier au lieu de mmoire en RAM pour excution)
	- Pas de dcompression des programmes ASM (et excution bloque)

*/




/******************************************************************************
*
* project name:    TIGCC Tools Suite
* file name:       ttstart.c
* initial date:    14/08/2000
* author:          thomas.nussbaumer@gmx.net
* description:     start utility for exepacked and/or normal executables
*
* =============================================================================
* IMPORTANT NOTE
* =============================================================================
*
* This program uses many really dirty tricks to overcome some restrictions of
* the AMS. Don't take it as example how to code (dirty tricks are NOT smart!)
*
* The special RETURN_VALUE treatment exit is, of course, the most dirty part.
* It's correctnes can ONLY be guaranteed for TIGCC 0.93 with optimization
* setting -Os !!!!
*
* And: Don't modify the source when you haven't completely understood it!
*
*
* =============================================================================
* Usage info - normal build (tigcc -Os ttstart.c)
* =============================================================================
*
* When ttstart is build "normally" it can be used to start either exepacked or
* normal executables on TI-89/TI-92plus running AMS versions up to 2.05.
* To start a program (for example: myprog.ppg or myprog.ASM) just type:
* ttstart("myprog")
* (NOTE: the quotes around the program name!)
* If you want to hand-over some commandline parameters to the started program
* just specify them after the programname like this: ttstart("myprog",10,20)
* The program name itself is removed from the estack by ttstart. The started
* program will only see the additionally parameters of ttstart.
*
* Additionally:
*
* ttstart is the first available launcher which handles return values >>from the
* started program to the AMS correctly.
*
*
* =============================================================================
* customizing ttstart (generating own, special purpose launchers)
* =============================================================================
*
* -----------------------------------------------------------------------------
* (1) using a predefined filename
* -----------------------------------------------------------------------------
*
* if you want to build a launcher which automatically starts a predefined
* program you can use the TIGCC commandline option -DCUSTOM_NAME=name_of_file
* For example: if you want to build a launcher with name mylaunch which should
* automatically start myprog.ppg compile ttstart like this:
*
* tigcc -Os -DCUSTOM_NAME=myprog -o mylaunch
*
* After uploading mylaunch and myprog to your calc you can simply start myprog
* by just typing: mylaunch()
* You can also hand-over any parameter to myprog. For example: mylaunch(10,20)
* starts myprog with parameter 10 and 20 on the estack.
*
* -----------------------------------------------------------------------------
* (2) customizing the type of program (normal or exepacked)
* -----------------------------------------------------------------------------
*
* It doesn't make much sense to build a launcher with a predefined filename,
* but without customizing the type of the program, too.
* If you build a launcher for your program and the program is NOT exepacked,
* the launcher will still contain the unpacking code (around 1200 bytes) which
* is not necessary. Therefore you can use commandline option -DONLY_PPG
* for exepacked programs or -DONLY_ASM for normal program.
*
* -------- Example 1 - launcher for myprog.ppg --------
*
* tigcc -Os -DCUSTOM_NAME=myprog -DONLY_PPG -o mylaunch
*
* -------- Example 2 - launcher for myprog.asm --------
*
* tigcc -Os -DCUSTOM_NAME=myprog -DONLY_ASM -o mylaunch
*
* -----------------------------------------------------------------------------
* (3) memory leak detection support (-DLEAKWATCH)
* -----------------------------------------------------------------------------
*
* When you compile ttstart with option -DLEAKWATCH it will check if the
* available memory gets changed by the started program. If the memory is changed
* it will output the difference in the status line (negative number == less
* memory available after program run / positive number == more memory available
* after program run).
*
* Not each report is a real memory leak of the started program. If a program
* adds or deletes variables, or if it manipulates the home screen by adding
* or removing items from it, a change of available memory  will be, of course,
* reported by ttstart, too. There are several other possibilities for "legal"
* changes of the available memory, but if your program ONLY uses malloc()'s and
* free()'s it's very likely that you have a real memory leak.
*
*
* =============================================================================
* TIGCC -pack option (how to make a new pstarter.o file)
* =============================================================================
*
* Since version 0.90 TIGCC features option -pack which generates automatically
* exepacked programs including a custom launcher program.
* The object code of the custom launcher (pstarter.o) used by TIGCC can be
* found in directory include/c of your TIGCC installation.
* pstarter.o can be easily generated from ttstart.c:
*
* [since TIGCC 0.93 no longer an assembly file but an object file is used]
*
* (1) compile ttstart.c this way:
*     tigcc -c -Os -DCUSTOM_NAME=tempprog -DONLY_PPG ttstart.c
* (2) rename generated ASM file ttstart.o into pstarter.o
* (3) backup old pstarter.o from directory include/c of your TIGCC
*     installation
* (4) copy the newly generated file pstarter.o over the old one
*     in directory include/c of your TIGCC installation
*
*
* =============================================================================
* How to generate exepacked programs
* =============================================================================
*
* The simplest way to generate an exepack program is, of course, the -pack
* option of TIGCC (examine the TIGCC docs).
* The second way to generate an exepack program from ANY executable (even from
* kernel-dependent programs) is to use tool ttppggen from the TIGCC Tools
* Suite.
*
*
* =============================================================================
* CREDITS
* =============================================================================
*
* (1) Zeljko Juric, for the "normal launcher" code from the TIGCC-FAQ and the
*     code which overcomes the RETURN_VALUE problem
* (2) Kevin Kofler, for the idea how to handle RETURN_VALUE in the context of a
*     launcher correctly, for the new version of own_enter_ghost_space and
*     other modifications related to TIGCC 0.93
* (3) Greg Dietsche, for various suggestions (normal program handling etc.)
* (4) Sebastian Reichelt, for fixing the comments
* (5) Olivier Armand (ExtendeD) for the necessary modifications to support
*     kernel-dependent exepacked prgs and for the EX_patch replacement
* (6) Me (Thomas Nussbaumer), for anything else
*
* $Id: ttstart.c,v 1.17 2002/05/27 07:54:09 tnussb Exp $
*
******************************************************************************/

#define USE_TI89       // compile for TI89
#define USE_TI92P      // compile for TI92p
#define USE_V200       // compile for V200

#define MIN_AMS 100         // works on all AMS versions
#define NO_EXIT_SUPPORT     // the launcher doesn't need exit support
#define NO_CALC_DETECT      // it doesn't need calc detection, either
#define OPTIMIZE_ROM_CALLS  // generates smaller code by reserving a5

//-----------------------------------------------------------------------------
// DIRTY: This is needed to get everything to work with "complex main"s
//        (Kevin Kofler)
//-----------------------------------------------------------------------------
asm("/* Include Patch: save_the_sp */");
extern void *__save__sp__;

#include <statline.h>
#include <alloc.h>
#include <vat.h>
#include <string.h>
#include <system.h>
#include <error.h>

#ifndef CUSTOM_NAME
#include <args.h>
#endif

//-----------------------------------------------------------------------------
// the following variable is used to protect a launcher from launching itself
// NOTE, that this is only enabled for the commandline version and NOT for
// customized launchers, because I think that a programmer KNOWS what he is
// doing and building a customized launcher to launch itself doesn't make
// ANY sense!
//-----------------------------------------------------------------------------


short its_me_checking = 0;

#define fatal(s)         {its_me_checking=0; ST_showHelp(s); return;}
#define fatal_without(s) {ST_showHelp(s); return;}

#include "ttunpack.h"    // errorcodes definition

//-----------------------------------------------------------------------------
// embedded version string -- EDITED BY HAND!
//
// NOTE: this version string is embedded in ALL variants and its should !!NOT!!
//       be removed by anyone
//
// to force the compiler to embed the string in all case this variable is
// volatile by intention!
//-----------------------------------------------------------------------------
volatile const char embedded_ttstart_version[]="[tts117]";


//=============================================================================
// nothing special, just a huge main routine ...
//=============================================================================
void _main(void) {
    unsigned char*         src;
    unsigned char*         dest;
    unsigned short         length;
    char                   tmpstr[50];
    const char*            fname;
    SYM_ENTRY*             symptr, *sym;
    HANDLE                 h;
    // check if "we" are already active (the same copy of this executable) and
    // if yes exit immediately !!
    if (its_me_checking) {fatal_without("can't start myself");}
    else                 its_me_checking = 1;  // now we are active ...
#ifdef LEAKWATCH
    long freemem = (long)HeapAvail();
#endif
    //-------------------------------------------------------------------------
    // Here comes the first "DIRTY, DIRTY, DONT-DO-IT-NORMALLY" hack:
    //
    // We are not using a copy of the top_estack pointer, but we operate
    // directly on the top_estack pointer itself. This way the called program
    // will not "see" its own program name on the estack !!!!
    //
    // NOTE: due to the "nature" of the estack it is not necessary to restore
    //       the top_estack pointer to its original value
    //-------------------------------------------------------------------------
    if (GetArgType(top_estack) != STR_TAG)
    {
        fatal("use: ttstart(\"name\")");
    }
    fname = GetStrnArg(top_estack);
    // suggestion from Greg Dietsche: if given filename contains a '(' truncate
    // the string there. This becomes useful if someone uses a auto-completion
    // TSR
    if ((src = strchr(fname,'(')))
        *src = 0;
    //-------------------------------------------------------------------------
    // setup TIOS name for variable and try to find it in the current folder ..
    //-------------------------------------------------------------------------
    tmpstr[0] = 0;
    strcpy(tmpstr+1,fname);
    if (!(symptr = DerefSym(SymFind(tmpstr+1+strlen(fname)))))
    {
        sprintf(tmpstr,"%s not found",fname);
        fatal(tmpstr);
    }
    //-------------------------------------------------------------------------
    // IMPORTANT: only use handle (symptr may become invalid !!!)
    //-------------------------------------------------------------------------
    h = symptr->handle;
    //-------------------------------------------------------------------------
    // if file is already locked, don't lock it again ...
    //-------------------------------------------------------------------------
    if (HeapGetLock(h))
    {
        src = HeapDeref(h);
        h = 0;
    }
    else
    {
        src = HLock(h);
    }
    if (!src)
        fatal("locking failed"); // will return!
    length = *(unsigned short *)src;
    dest = src+length+1;
    //-------------------------------------------------------------------------
    // check if given program is not even a packed file
    //-------------------------------------------------------------------------
    //---------------------------------------------------------------------
    // skip "length of variable" bytes
    //---------------------------------------------------------------------
    src += 2;
    if (*dest != 0xF8 || *(dest-4) != 'p' || *(dest-3) != 'p' || *(dest-2) != 'g' || !ttunpack_valid(src))
    {
        if (h)
            HeapUnlock(h);
        fatal("Not a PPG program");
    }
    //---------------------------------------------------------------------
    // get the original size and allocate memory for it
    //---------------------------------------------------------------------
    length = ttunpack_size(src);
    sym = DerefSym (SymAdd ($(unpacked)));
    if (!(dest = (char*)HeapDeref (sym->handle = HeapAlloc(length))))
    {
        if (h)
            HeapUnlock(h);
        SymDel (tmpstr+1+strlen(fname));
        fatal("out of memory");
    }
    ST_showHelp("decompressing ...");
    if (ttunpack_decompress(src,dest))
    {
        //-------------------------------------------------------------
        // "houston we got problems ..."
        //-------------------------------------------------------------
        if (h) HeapUnlock(h);
        SymDel (tmpstr+1+strlen(fname));
        fatal("failed");
    }
    ST_eraseHelp();

    // Inutile d'excuter -> on supprime le code d'excution

    if (h)
        HeapUnlock(h);
#ifdef LEAKWATCH
    if ((freemem=((long)HeapAvail())-freemem) != 0) {
        sprintf(tmpstr, "mem changed by %ld", freemem);
        ST_showHelp(tmpstr);
    }
#endif
    its_me_checking = 0; // we are not active anymore ...
}

//#############################################################################
//###################### NO MORE FAKES BEYOND THIS LINE #######################
//#############################################################################
//
//=============================================================================
// Revision History
//=============================================================================
//
// $Log: ttstart.c,v $
// Revision 1.17  2002/05/27 07:54:09  tnussb
// exception handling added (thanx to Sebastian Reichelt)
//
// Revision 1.16  2002/05/22 09:01:38  tnussb
// embedded version string set to tts116 to reflect the "stepping back to old
// unpacking function" modification
//
// Revision 1.15  2002/04/17 10:35:10  tnussb
// documentation extended
//
// Revision 1.14  2002/03/13 15:01:42  tnussb
// (1) some messages shortened
// (2) using now new and minor size optimized unpacking function
// (3) some comments added
//
// Revision 1.13  2002/03/04 08:51:03  tnussb
// Extended's relocation routine as smart ASM macro added (just for the case
// that EX_patch() will "break" again)
//
// Revision 1.12  2002/02/28 08:53:18  tnussb
// EX_patch called fixed to be prepared for new EX_patch version coming with
// AMS 2.06
//
// Revision 1.11  2002/02/07 09:57:58  tnussb
// many changes like new bootstrapping, new enter_ghost_space routine and,
// of course, kernel-dependend PPG support
//
// Revision 1.9  2001/07/31 22:56:59  Thomas Nussbaumer
// generic commit for Tools Suite version 0.99.10
//
// Revision 1.8  2001/02/04 16:45:40  Thomas Nussbaumer
// using new ttunpack.h version
//
// Revision 1.7  2000/11/18 16:00:18  Thomas Nussbaumer
// enter_ghost_space() added
//
// Revision 1.6  2000/10/01 14:56:52  Thomas Nussbaumer
// support for custom start utility build added
//
// Revision 1.5  2000/08/30 20:01:55  Thomas Nussbaumer
// using now assembler optimized version of unpack routine
//
// Revision 1.4  2000/08/30 19:44:24  Thomas Nussbaumer
// TABs replaced by spaces and trailing spaces removed
//
// Revision 1.3  2000/08/20 15:28:22  Thomas Nussbaumer
// including of unpack.c done now before including ttunpack.h
//
// Revision 1.2  2000/08/16 22:29:05  Thomas Nussbaumer
// (1) based now on shared file unpack.c
// (2) argument can be specified on commandline
// (3) define and code for time measurement integrated
//
// Revision 1.1  2000/08/14 22:48:38  Thomas Nussbaumer
// initial version
//
//


